import unittest
from ts_tokenizer.token_handler import TokenPreProcess


class TestTokenPreProcess(unittest.TestCase):

    def test_is_xml(self):
        words = ['<text id="001">', "<test>", "</text>"]
        expected_results = ['<text id="001">', None, '</text>']
        for word, expected in zip(words, expected_results):
            result = TokenPreProcess.is_xml(word)
            if result:
                result = result[0]  # Extract the XML tag string from the tuple
            self.assertEqual(result, expected)

    def test_is_mention(self):
        words = ['@tanerim', 'user', '@tscorpus', 'invalid']
        expected_results = [('@tanerim', 'Mention'), None, ('@tscorpus', 'Mention'), None]

        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_mention(word), expected)

    def test_is_hashtag(self):
        words = ['#hashtag', 'nohashtag', '#12345', '#invalid!']
        expected_results = [('#hashtag', 'Hashtag'), None, ('#12345', 'Hashtag'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_hashtag(word), expected)

    def test_is_in_quotes(self):
        words = ['"araba"', "'araba'", 'Araba']
        expected_results = [[('"', 'Punc'), ('araba', 'Valid_Word'), ('"', 'Punc')],
                            [("'", 'Punc'), ('araba', 'Valid_Word'), ("'", 'Punc')],
                            None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_in_quotes(word), expected)

    def test_is_in_parenthesis(self):
        words = ['(parantez)', '(parantez', 'no)']
        expected_results = [[('(', 'Punc'), ('parantez', 'Valid_Word'), (')', 'Punc')], None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_in_parenthesis(word), expected)

    def test_is_hour(self):
        words = ['12:30', '25:00', '5:99']
        expected_results = [('12:30', 'Hour'), None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_hour(word), expected)

    def test_is_date(self):
        words = ['12/05/2023', '31.02.2020', '32.14.2024']
        expected_results = [('12/05/2023', 'Date'), None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_date(word), expected)

    def test_is_percentage_numbers(self):
        words = ['%50', '50%', '%abc']
        expected_results = [('%50', 'Percentage_Numbers'), ('50%', 'Percentage_Numbers'), None]
        for word, expected in zip(words, expected_results):
            if isinstance(expected, list):  # For cases where multiple tokens are returned
                self.assertEqual(TokenPreProcess.is_percentage_numbers_chars(word), expected)
            else:  # For single token results
                self.assertEqual(TokenPreProcess.is_percentage_numbers(word), expected)

    def test_is_percentage_numbers_chars(self):
        words = ['%50USD', 'USD%50',]
        expected_results = [[('%50', 'Percentage_Numbers'), ('USD', 'Abbr')], None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_percentage_numbers_chars(word), expected)

    def test_is_roman_number(self):
        words = ['X', 'V', '123', 'invalid']
        expected_results = [('X', 'Roman_Number'), ('V', 'Roman_Number'), None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_roman_number(word), expected)

    def test_is_email_punc(self):
        words = ['.test@example.com.', 'test@example.com.', 'invalid']
        expected_results = [
            [('.', 'Punc'), ('test@example.com', 'Email'), ('.', 'Punc')],
            [('test@example.com', 'Email'), ('.', 'Punc')],
            None
        ]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_email_punc(word), expected)

    def test_is_email(self):
        words = ['test@example.com', 'user@domain', 'invalid_email']
        expected_results = [('test@example.com', 'Email'), None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_email(word), expected)

    def test_is_url(self):
        words = ['http://example.com', 'www.test.com', 'not_a_url', 'https://example.com',]
        expected_results = [('http://example.com', 'URL'), ('www.test.com', 'URL'), None, ('https://example.com', 'URL')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_url(word), expected)

    def test_is_copyright(self):
        words = ['©2023', '2023©', 'invalid']
        expected_results = [('©2023', 'Copyright'), ('2023©', 'Copyright'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_copyright(word), expected)

    def test_is_registered(self):
        words = ['®Ibanez', 'Fender®', 'invalid']
        expected_results = [('®Ibanez', 'Registered'), ('Fender®', 'Registered'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_registered(word), expected)

    def test_is_currency(self):
        words = ['$100', '100€', 'invalid']
        expected_results = [('$100', 'Currency'), ('100€', 'Currency'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_currency(word), expected)

    def test_is_abbr(self):
        words = ['vb', 'TBMM', 'invalid']
        expected_results = [('vb', 'Abbr'), ('TBMM', 'Abbr'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_abbr(word), expected)

    def test_is_in_lexicon(self):
        words = ['araba', 'ilamklamuli', 'arabalarındaki']
        expected_results = [('araba', 'Valid_Word'), None, ('arabalarındaki', 'Valid_Word')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_in_lexicon(word), expected)

    def test_is_number(self):
        words = ['12345', 'invalid', '123.456', '123,456']
        expected_results = [('12345', 'Number'), None, None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_number(word), expected)

    def test_is_in_exceptions(self):
        words = ['dvd-rom', 'notexception', 'sosyo-demografik']
        expected_results = [('dvd-rom', 'Exception'), None, ('sosyo-demografik', 'Exception')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_in_exceptions(word), expected)

    def test_is_in_eng_words(self):
        words = ['hello', 'notenglish', 'car']
        expected_results = [('hello', 'English_Word'), None, ('car', 'English_Word')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_in_eng_words(word), expected)

    def test_is_smiley(self):
        words = [':)', ':D', 'invalid']
        expected_results = [(':)', 'Smiley'), (':D', 'Smiley'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_smiley(word), expected)

    def test_is_multiple_smiley(self):
        words = [':):)', ':-)', 'invalid']
        expected_results = [(':):)', 'Multiple_Smiley'), None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_multiple_smiley(word), expected)

    def test_is_multiple_smiley_in(self):
        words = ['Heyyo:-):-)', ':-)', 'invalid', 'yes:):)']
        expected_results = [('Heyyo:-):-)', 'Multiple_Smiley_In'), None, None, ('yes:):)', 'Multiple_Smiley_In')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_multiple_smiley_in(word), expected)

    def test_is_emoticon(self):
        words = ['🖤', ':(', '🥰']
        expected_results = [('🖤', 'Emoticon'), None, ('🥰', 'Emoticon')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_emoticon(word), expected)

    def test_is_multiple_emoticon(self):
        words = ['🥰🥰', '🥰', '🖤🥰']
        expected_results = [('🥰🥰', 'Multiple_Emoticon'),  None, ('🖤🥰', 'Multiple_Emoticon')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_multiple_emoticon(word), expected)

    def test_is_fsp(self):
        words = ['araba.']
        expected_results = [[('araba', 'Valid_Word'), ('.', 'Punc')]]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_fsp(word), expected)

    def test_is_isp(self):
        words = ['.araba']
        expected_results = [[('.', 'Punc'), ('araba', 'Valid_Word')]]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_isp(word), expected)

    def test_is_mssp(self):
        words = ['.araba.', '..araba..', '.']
        expected_results = [[('.', 'Punc'), ('araba', 'Valid_Word'), ('.', 'Punc')], None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_mssp(word), expected)

    def test_is_msp(self):
        words = ['...araba...', 'araba']
        expected_results = [[('...', 'Punc'), ('araba', 'Valid_Word'), ('...', 'Punc')], None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_msp(word), expected)

    def test_is_imp(self):
        words = ['...araba', 'araba...']
        expected_results = [[('...', 'Punc'), ('araba', 'Valid_Word')], None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_imp(word), expected)

    def test_is_fmp(self):
        words = ['araba...', '...araba', 'Geldim(!)']
        expected_results = [[('araba', 'Valid_Word'), ('...', 'Punc')], None, [('Geldim', 'Valid_Word'), ('(!)', 'Punc')]]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_fmp(word), expected)

    def test_is_apostrophed(self):
        words = ["Defne'nin", "Bengü\"nün", "Türkiye ́de"]
        expected_results = [("Defne'nin", 'Apostrophed'), None, ("Türkiye'de", 'Apostrophed')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_apostrophed(word), expected)

    def test_is_non_latin(self):
        words = ['你好', '𝐆eldim']
        expected_results = [('你好', 'Non_Latin'), ('𝐆eldim', 'Non_Latin')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_non_latin(word), expected)

    def test_is_one_char_fixable(self):
        words = ['gelmek¬te', 'valid', 'bili-yorum']
        expected_results = [('gelmekte', 'One_Char_Fixed'), None, ('biliyorum', 'One_Char_Fixed')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_one_char_fixable(word), expected)

    def test_is_punc(self):
        words = ['!!!', 'invalid!', '(!)', '---']
        expected_results = [('!!!', 'Punc'), None, ('(!)', 'Punc'), ('---', 'Punc')]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_punc(word), expected)

    def test_is_underscored(self):
        words = ['word_123', '_yeni_', 'geçersiz_', '_geçersiz']
        expected_results = [('word_123', 'Alphanumeric_Underscored'), None, None, None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_underscored(word), expected)

    def test_is_three_or_more(self):
        words = ['-----', '.....', '!!']
        expected_results = [('-----', 'Three_Or_More'), ('.....', 'Three_Or_More'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_three_or_more(word), expected)

    def test_is_date_range(self):
        words = ['1990-1995', '1995-2000', '(invalid-range)']
        expected_results = [('1990-1995', 'Date_Range'), ('1995-2000', 'Date_Range'), None]
        for word, expected in zip(words, expected_results):
            self.assertEqual(TokenPreProcess.is_date_range(word), expected)


if __name__ == '__main__':
    unittest.main()



import re
import string
from .data import LocalData
from .char_fix import CharFix
from .date_check import DateCheck
from .smiley_check import SmileyParser
from .emoticon_check import EmoticonParser
from .punctuation_process import PuncMatcher

puncs = re.escape(string.punctuation)
extra_puncs = ["–", "°", "—", "(", ")"]
puncs += re.escape(''.join(extra_puncs))
domains_pattern = '|'.join([re.escape(domain[1:]) for domain in LocalData.domains()])  # Escaping only necessary parts

# Create a dict of RegExps
REGEX_PATTERNS = {
    # Precompiled regular expressions using re.compile()
    "xml_tag": re.compile(r"<\s*\w+(\s+\w+\s*=\s*\"[^\"]+\")*\s*/?>|</\w+\s*>"),
    "hashtag": re.compile(r'^#[^#]{1,143}$'),
    "mention": re.compile(r'^@[^@]{1,143}$'),
    "email": re.compile(r'\b[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+\b(?![.,!?;:])'),
    "email_punc": re.compile(r'\b[' + re.escape(string.punctuation) + r']*[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+[' + re.escape(string.punctuation) + r']*\b'),
    "url_pattern": re.compile(fr'(?:(?:http|https|ftp)://)?(?:www\.)?[-a-zA-Z0-9:%._\\+~#=]{{1,256}}(?:{domains_pattern})(?:\.[a-zA-Z]{{2,3}})?(?:/[-a-zA-Z0-9()@:%_\\+.~#?&//=]*)?\b(?![.,!?;:])'),
    "hour": re.compile(r"\b(0[0-9]|1[0-9]|2[0-3])[:.][0-5][0-9](?: ?[AP]M)?(?:'te|'de|'da|'den|'dan|'ten|'tan|'deki|'daki)?(?=$|\s)"),
    "percentage_numbers_chars": re.compile(r'%(\d+\D+)'),
    "percentage_numbers": re.compile(r'(%\d{1,3}(?:\.\d{3})*(?:,\d+)?|\d{1,3}(?:\.\d{3})*(?:,\d+)?%)'),
    "single_hyphen": re.compile(r'^(?!-)[\w]+-[\w]+(?!-)$'),
    "date_range": re.compile(r'^\d{4}-\d{4}$'),
    "in_parenthesis": re.compile(r'^[\(\[\{][^()\[\]{}]*[\)\]\}]$'),
    "in_quotes": re.compile(r'^[\'"][^\'"]*[\'"]$'),
    "copyright": re.compile(r'(?:^©[a-zA-Z0-9]+$)|(?:^[a-zA-Z0-9]+©$)'),
    "registered": re.compile(r'(?:^®[a-zA-Z]+$)|(?:^[a-zA-Z]+®$)'),
    "initial_parenthesis": re.compile(r"^([\(\[\{]+)([^\)]+)([\)\]\}]+)(.*)$"),
    "three_or_more": re.compile(r'^([{}])\1{{2,}}$'.format(re.escape(string.punctuation))),
    "num_char_sequence": re.compile(r'\d+[\w\s]*'),
    "roman_number": re.compile(r'^(M{0,4}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3}))\.?$'),
    "apostrophed": re.compile(r"\b\w+'[a-zA-ZıiİüÜçÇöÖşŞğĞ]+\b"),
    "currency": re.compile(rf"^(?:[{re.escape(''.join(LocalData.currency_symbols()))}]\d{{1,3}}(?:[.,]\d{{3}})*([.,]\d+)?|\d{{1,3}}(?:[.,]\d{{3}})*([.,]\d+)?[{re.escape(''.join(LocalData.currency_symbols()))}])$")
}


exception_list = ["(!)", "...", "[...]"]


#def check_regex(word, pattern):
#    return word if re.search(REGEX_PATTERNS[pattern], CharFix.fix(word)) else None

def check_regex(word, pattern):
    return word if REGEX_PATTERNS[pattern].search(CharFix.fix(word)) else None


def punc_count(word: str) -> int:
    return sum(1 for char in word if char in puncs)


def punc_pos(word: str) -> list:
    return [i for i, char in enumerate(word) if char in puncs]


def split_on_punctuation(word: str, punctuation_list: str) -> list:
    result = []
    i = 0
    for j, char in enumerate(word):
        if char in punctuation_list:
            if i < j:
                processed_token = TokenProcessor.process_token(word[i:j])
                result.append((word[i:j], processed_token[1]))
            result.append((char, "Punc"))
            i = j + 1
    if i < len(word):
        processed_token = TokenProcessor.process_token(word[i:j])
        result.append((word[i:j], processed_token[1]))
    return result


def apply_charfix(func):
    def wrapper(word, *args, **kwargs):
        fixed_word = CharFix.fix(word)
        return func(fixed_word, *args, **kwargs)
    return wrapper


def tr_lowercase(func):
    def wrapper(word, *args, **kwargs):
        turkish_lowercase = CharFix.tr_lowercase(word)
        return func(word, turkish_lowercase, *args, **kwargs)
    return wrapper


class TokenPreProcess:

    def __init__(self):
        pass

    # Regex Based Tokens
    # These functions get the input token, checks against to regexs defined above and
    # return word, tag as tuple
    @staticmethod
    def is_xml(word: str) -> tuple:
        # Check if the word starts with '<' and ends with '>', and there are no other '<' or '>' inside the word
        if word.startswith("<") and word.endswith(">") and (word.count("<") == 1 and word.count(">") == 1):
            if "/" not in word and " " in word:
                result = check_regex(word, "xml_tag")
                return (result, "XML_Tag") if result else None
            elif " " not in word and "/" in word:
                result = check_regex(word, "xml_tag")
                return (result, "XML_Tag") if result else None

    @staticmethod
    @apply_charfix
    def is_mention(word: str) -> tuple:
        p_count = PuncMatcher.punc_count(word)
        result = check_regex(word, "mention") if p_count == 1 else None
        return (result, "Mention") if result else None

    @staticmethod
    @apply_charfix
    def is_hashtag(word: str) -> tuple:
        p_count = PuncMatcher.punc_count(word)
        result = check_regex(word, "hashtag") if p_count == 1 else None
        return (result, "Hashtag") if result else None

    @staticmethod
    @apply_charfix
    def is_in_quotes(word: str) -> list:
        result = check_regex(word, "in_quotes")
        if result:
            initial_parenthesis = word[0]
            final_parenthesis = word[-1]
            content = word[1:-1]
            processed_content = TokenProcessor.process_token(content)
            if isinstance(processed_content, tuple):
                processed_content = [processed_content]
            return [(initial_parenthesis, "Punc")] + processed_content + [(final_parenthesis, "Punc")]

    @staticmethod
    @apply_charfix
    def is_in_parenthesis(word: str) -> list:
        if len(word) > 2:
            result = check_regex(word, "in_parenthesis")
            if result:
                initial_parenthesis = word[0]
                final_parenthesis = word[-1]
                content = word[1:-1]
                processed_content = TokenProcessor.process_token(content)
                if isinstance(processed_content, tuple):
                    processed_content = [processed_content]
                return [(initial_parenthesis, "Punc")] + processed_content + [(final_parenthesis, "Punc")]

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_initial_parenthesis(word: str, lower_word: str) -> list:
        result = check_regex(word, "initial_parenthesis")
        if result:
            parts = re.split(r"[)\]\}]", word)
            if word[0] == "(":
                parenthesis_content = parts[0] + ")"
            elif word[0] == "[":
                parenthesis_content = parts[0] + "]"
            elif word[0] == "{":
                parenthesis_content = parts[0] + "}"
            if re.match(r'^[\(\[\{]\d{1,2}[\)\]\}]$', parenthesis_content):  # Matches "(1)", "(20)", etc.
                numbered_title = [(parenthesis_content, "Numbered_Title")]
                remaining_part = parts[1].strip()
                processed_remaining = TokenProcessor.process_token(remaining_part) if remaining_part else []
                if isinstance(processed_remaining, tuple):
                    processed_remaining = [processed_remaining]
                return numbered_title + processed_remaining

            processed_parenthesis = TokenPreProcess.is_in_parenthesis(parenthesis_content)
            remaining_part = parts[1].strip()
            processed_remaining = TokenProcessor.process_token(remaining_part) if remaining_part else []
            if isinstance(processed_remaining, tuple):
                processed_remaining = [processed_remaining]

            # Ensure processed_parenthesis is not None
            return (processed_parenthesis or []) + processed_remaining

    @staticmethod
    def is_date_range(word: str) -> tuple:
        result = check_regex(word, "date_range")
        if result:
            return (result, "Date_Range") if result else None

    @staticmethod
    def is_date(word: str) -> tuple:
        result = DateCheck.is_date(word)
        return (word, "Date") if result else None

    @staticmethod
    def is_hour(word: str) -> tuple:
        result = check_regex(word, "hour")
        return (result, "Hour") if result else None

    @staticmethod
    def is_percentage_numbers(word: str) -> tuple:
        p_count = PuncMatcher.punc_count(word)
        result = check_regex(word, "percentage_numbers") if p_count == 1 else None
        if result:
            return (word, "Percentage_Numbers")

    @staticmethod
    @apply_charfix
    def is_percentage_numbers_chars(word: str) -> list:
        result = check_regex(word, "percentage_numbers_chars")
        if result:
            pattern = r'(%\d+(?:[.,]\d+)?(?:-\d+)?|\d+(?:[.,]\d+)?%|[a-zA-ZşŞıİçÇğĞöÖüÜ]+|[%\W]+)'
            tokens = re.findall(pattern, word)
            processed_tokens = []

            if tokens:
                first_token = tokens[0]
                if re.match(r'%\d+(?:[.,]\d+)?(?:-\d+)?|\d+(?:[.,]\d+)?%', first_token):
                    processed_tokens.append((first_token, "Percentage_Numbers"))
                else:
                    processed_tokens.append(TokenProcessor.process_token(first_token))

                for token in tokens[1:]:
                    processed_tokens.append(TokenProcessor.process_token(token))

            return processed_tokens if processed_tokens else [result, "Percentage_Numbers_Chars"]

    @staticmethod
    def is_roman_number(word: str) -> tuple:
        result = check_regex(word, "roman_number")
        return (result, "Roman_Number") if result else None

    @staticmethod
    def is_email_punc(word: str) -> list:
        result = check_regex(word, "email_punc")
        if result:
            start_punc_count = 0
            end_punc_count = 0
            for char in word:
                if char in puncs:
                    start_punc_count += 1
                else:
                    break
            for char in word[::-1]:
                if char in puncs:
                    end_punc_count += 1
                else:
                    break
            initial_punc = word[:start_punc_count] if start_punc_count > 0 else ""
            final_punc = word[-end_punc_count:] if end_punc_count > 0 else ""
            if start_punc_count > 0 and end_punc_count > 0:
                email_part = word[start_punc_count: -end_punc_count]
            elif start_punc_count > 0:
                email_part = word[start_punc_count:]
            elif end_punc_count > 0:
                email_part = word[:-end_punc_count]
            else:
                email_part = word
            result_list = []
            if initial_punc:
                result_list.append((initial_punc, "Punc"))
            if email_part:
                result_list.append((email_part, "Email"))
            if final_punc:
                result_list.append((final_punc, "Punc"))
            return result_list

    @staticmethod
    def is_email(word: str) -> tuple:
        result = check_regex(word, "email")
        if result and any(dne in word for dne in LocalData.domains()) and word[0] not in puncs and word[-1] not in puncs:
            return (result, "Email") if result else None

    @staticmethod
    def is_url(word: str) -> tuple:
        if any(dne in word for dne in LocalData.domains()) and "@" not in word and word[0] not in puncs and word[-1] not in [")", "(", "[", "]"]:
            result = check_regex(word, "url_pattern")
            if "'" in word:
                return (result, "URL_Suffix") if result else None
            else:
                return (result, "URL") if result else None

    @staticmethod
    def is_copyright(word: str) -> tuple:
        result = check_regex(word, "copyright")
        return (result, "Copyright") if result else None

    @staticmethod
    def is_registered(word: str) -> tuple:
        result = check_regex(word, "registered")
        return (result, "Registered") if result else None

    @staticmethod
    def is_currency(word: str) -> tuple:
        result = check_regex(word, "currency")
        return (result, "Currency") if result else None

    @staticmethod
    @apply_charfix
    def is_num_char_sequence(word: str) -> tuple:
        result = check_regex(word, "num_char_sequence")
        return (result, "Num_Char_Sequence") if result else None

    # Lexicon Based Tokens
    @staticmethod
    @apply_charfix
    def is_abbr(word: str) -> tuple:
        # Use lower_word for comparison
        return (word, "Abbr") if word in LocalData.abbrs() else None

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_in_lexicon(word: str, lower_word: str) -> tuple:
        if lower_word in LocalData.word_list():
            return word, "Valid_Word"

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_in_exceptions(word: str, lower_word: str) -> tuple:
        return (word, "Exception") if lower_word in LocalData.exception_words() else None

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_in_eng_words(word, lower_word: str):
        word_fixed = CharFix.fix(word)
        if lower_word in LocalData.eng_word_list():
            return word_fixed, "English_Word"
        return None

    @staticmethod
    def is_smiley(word):
        return (word, "Smiley") if word in LocalData.smileys() else None

    @staticmethod
    def is_emoticon(word):
        return (word, "Emoticon") if word in LocalData.emoticons() else None

    # Multi-Unit Tokens
    @staticmethod
    def is_multiple_smiley(word):
        if SmileyParser.consecutive_smiley(word) and not str(word[0:-1]).isalnum():
            return word, "Multiple_Smiley"
        return None

    @staticmethod
    def is_multiple_smiley_in(word):
        if SmileyParser.consecutive_smiley(word) and any(char.isalnum() for char in word):
            return word, "Multiple_Smiley_In"
        return None

    @staticmethod
    def is_multiple_emoticon(word):
        return (word, "Multiple_Emoticon") if EmoticonParser.emoticon_count(word) >= 2 else None

    @staticmethod
    def is_number(word):
        return (word, "Number") if word.isdigit() else None

    @staticmethod
    @apply_charfix
    def is_fsp(word: str) -> list:
        if len(word) > 1 and word[-1] in puncs and all(char not in puncs for char in word[:-1]):
            final_punc = word[-1]
            remaining_word = word[:-1]
            processed_word = TokenProcessor.process_token(remaining_word)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return processed_word + [(final_punc, "Punc")]

        elif len(word) > 1 and word[-1] in puncs and punc_count(word) > 2:
            final_punc = word[-1]
            remaining_word = word[:-1]
            processed_word = split_on_punctuation(remaining_word, puncs)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return processed_word + [(final_punc, "Punc")]

        elif word[-1] in puncs:
            final_punc = word[-1]
            remaining_word = word[:-1]
            processed_word = TokenProcessor.process_token(remaining_word)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return processed_word + [(final_punc, "Punc")]

    @staticmethod
    @apply_charfix
    def is_isp(word: str) -> list:
        if len(word) > 1 and word[0] in puncs and all(char not in puncs for char in word[1:]):
            initial_punc = word[0]
            remaining_word = word[1:]
            processed_word = TokenProcessor.process_token(remaining_word)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return [(initial_punc, "Punc")] + processed_word

        elif len(word) > 1 and word[0] in puncs and punc_count(word) > 2:
            initial_punc = word[0]
            remaining_word = word[1:]
            processed_word = split_on_punctuation(remaining_word, puncs)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return [(initial_punc, "Punc")] + processed_word

        elif word[0] in puncs:
            initial_punc = word[0]
            remaining_word = word[1:]
            processed_word = TokenProcessor.process_token(remaining_word)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return [(initial_punc, "Punc")] + processed_word


    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_mssp(word: str, lower_word: str) -> list:
        if len(word) > 2 and word[0] in puncs and word[-1] in puncs and all(char not in puncs for char in word[1:-1]):
            initial_punc = TokenProcessor.process_token(word[0])
            final_punc = TokenProcessor.process_token(word[-1])
            remaining_word = lower_word[1:-1]
            processed_word = TokenProcessor.process_token(remaining_word)
            if isinstance(processed_word, tuple):
                processed_word = [processed_word]
            return [initial_punc] + processed_word + [final_punc]

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_msp(word: str, lower_word: str) -> list:
        if len(word) > 2 and word not in exception_list:
            start_punc_count = 0
            end_punc_count = 0
            # Count starting punctuation
            for char in word:
                if char in puncs:
                    start_punc_count += 1
                else:
                    break
            # Count ending punctuation
            for char in word[::-1]:
                if char in puncs:
                    end_punc_count += 1
                else:
                    break
            # Ensure word has both starting and ending punctuations and valid middle part
            if start_punc_count >= 1 and end_punc_count >= 1 and all(
                    char not in puncs for char in word[start_punc_count: -end_punc_count]):
                initial_punc = word[:start_punc_count]
                final_punc = word[-end_punc_count:]
                remaining_word = lower_word[start_punc_count: -end_punc_count]
                if remaining_word == '':
                    return [TokenProcessor.process_token(initial_punc), TokenProcessor.process_token(final_punc)]
                processed_word = TokenProcessor.process_token(remaining_word)
                if isinstance(processed_word, tuple):
                    processed_word = [processed_word]
                return [TokenProcessor.process_token(initial_punc)] + processed_word + [TokenProcessor.process_token(final_punc)]

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_imp(word: str, lower_word: str) -> list:
        if len(word) > 1 and word not in exception_list:
            start_punc_count = 0
            for char in word:
                if char in puncs:
                    start_punc_count += 1
                else:
                    break
            if start_punc_count >= 2 and all(char not in puncs for char in word[start_punc_count:]):
                initial_punc = [(word[:start_punc_count], "Punc")]
                remaining_word = lower_word[start_punc_count:]
                processed_word = TokenProcessor.process_token(remaining_word)
                if isinstance(processed_word, tuple):
                    processed_word = [processed_word]
                return initial_punc + processed_word

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_fmp(word: str, lower_word: str) -> tuple:
        fmp_regex = re.compile(rf"[{puncs}]{{2,}}$")
        if fmp_regex.search(word):
            first_punc_index = next((i for i, char in enumerate(word) if char in puncs), None)
            before_punc = word[:first_punc_index]
            from_punc = word[first_punc_index:]
            if from_punc in exception_list:
                return [TokenProcessor.process_token(before_punc), (from_punc, "Punc")]
            if len(word) > 1 and word not in exception_list:
                end_punc_count = 0
                for char in word[::-1]:
                    if char in puncs:
                        end_punc_count += 1
                    else:
                        break
                if end_punc_count >= 2 and all(char not in puncs for char in word[:-end_punc_count]):
                    final_punc = word[-end_punc_count:]
                    remaining_word = word[:-end_punc_count]
                    lower_remaining_word = lower_word[:-end_punc_count]
                    processed_word = TokenProcessor.process_token(lower_remaining_word)
                    if isinstance(processed_word, tuple):
                        processed_word = [processed_word]
                    if processed_word:
                        processed_word[0] = (remaining_word, processed_word[0][1])
                    return processed_word + [(final_punc, "Punc")]

    @staticmethod
    @apply_charfix
    def is_apostrophed(word: str) -> tuple:
        result = check_regex(word, "apostrophed")
        if result:
            return word, "Apostrophed"

    @staticmethod
    @apply_charfix
    def is_midp(word: str) -> list:
        if len(word) > 2:
            if word[0] not in puncs and word[-1] not in puncs:
                mid_punc_pos = [i for i in range(1, len(word) - 1) if word[i] in puncs]
                if len(mid_punc_pos) == 1:
                    mid_punc_idx = mid_punc_pos[0]
                    initial_part = word[:mid_punc_idx]
                    mid_punc = word[mid_punc_idx]
                    remaining_part = word[mid_punc_idx + 1:]
                    processed_initial = TokenProcessor.process_token(initial_part)
                    processed_remaining = TokenProcessor.process_token(remaining_part)
                    if isinstance(processed_initial, tuple):
                        processed_initial = [processed_initial]
                    if isinstance(processed_remaining, tuple):
                        processed_remaining = [processed_remaining]
                    return processed_initial + [(mid_punc, "Punc")] + processed_remaining

    @staticmethod
    def is_punc(word):
        if word in exception_list:
            return word, "Punc"
        return (word, "Punc") if all(char in puncs for char in word) else None

    @staticmethod
    @apply_charfix
    @tr_lowercase
    def is_underscored(word: str, lower_word: str) -> tuple:
        if "_" in word and 1 <= word.count("_") <= 1 and word[0] != "_" and word[-1] != "_":
            parts = lower_word.split("_")
            # Check if it is alphanumeric underscored
            if len(parts) == 2 and parts[0].isalpha() and parts[1].isdigit() or parts[0].isdigit() and parts[1].isalpha():
                return word, "Alphanumeric_Underscored"
            # Check if all parts exist in the lexicon
            elif all(part in LocalData.word_list() for part in parts):
                return word, "Underscored"
            return word, "OOV"

    @staticmethod
    @apply_charfix
    def is_three_or_more(word):
        exceptions = ["...", "!!!"]
        if word in exceptions:
            return word, "Punc"
        result = check_regex(word, "three_or_more")
        return (word, "Three_Or_More") if result else None

    @staticmethod
    def verify_latin_chars(word: str) -> bool:
        return all('a' <= char <= 'z' or 'A' <= char <= 'Z' or char in "ğüşıöçĞÜŞİÖÇ'" for char in word)

    @staticmethod
    @apply_charfix
    def is_non_latin(word):
        allowed_chars = set("abcçdefgğhıijklmnoöprsştuüvyzwqxâîûABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZWQXÂÎ")
        sum_foreign_char = sum(
            1 for char in word if char not in allowed_chars and char not in puncs and not char.isdigit())
        sum_punc = PuncMatcher.punc_count(word)
        has_digit = any(char.isdigit() for char in word)
        hyphen_check = PuncMatcher.hyphen_in(word)
        underscore_check = TokenPreProcess.is_underscored(word)
        multiple_emoticon = TokenPreProcess.is_multiple_emoticon(word)
        if sum_foreign_char >= 1 and sum_punc == 0 and not has_digit and not hyphen_check and not multiple_emoticon and not underscore_check:
            return word, "Non_Latin"
        return None

    @staticmethod
    def is_one_char_fixable(word):
        extra_chars = ["¬", "-", "º", "0", "1"]
        for extra in extra_chars:
            if PuncMatcher.punc_pos(extra) != [0] or PuncMatcher.punc_pos(word) != [-1]:
                fixed_word = word.replace(extra, "")
                if TokenPreProcess.is_in_lexicon(fixed_word):
                    return fixed_word, "One_Char_Fixed"
        return None


check_methods = [
        # Check for SGML tags
        TokenPreProcess.is_xml,

        # First check is its is a smiley
        TokenPreProcess.is_smiley,

        # Then check if input is only composed of punctuations.
        TokenPreProcess.is_punc,

        # Lexicon Based Tokens
        TokenPreProcess.is_abbr,
        TokenPreProcess.is_in_exceptions,
        TokenPreProcess.is_in_lexicon,
        TokenPreProcess.is_in_eng_words,
        TokenPreProcess.is_emoticon,
        TokenPreProcess.is_number,

        # Specific Cases for Punctuation Use
        TokenPreProcess.is_in_quotes,
        TokenPreProcess.is_in_parenthesis,
        TokenPreProcess.is_initial_parenthesis,

        # Before validation make sure it is apostrophed
        TokenPreProcess.is_apostrophed,
        TokenPreProcess.is_underscored,
        TokenPreProcess.is_email,
        TokenPreProcess.is_email_punc,
        TokenPreProcess.is_url,
        TokenPreProcess.is_date_range,
        TokenPreProcess.is_date,
        TokenPreProcess.is_hour,

        # Basic Regex Tokens
        TokenPreProcess.is_mention,
        TokenPreProcess.is_hashtag,
        TokenPreProcess.is_copyright,
        TokenPreProcess.is_registered,
        TokenPreProcess.is_currency,
        TokenPreProcess.is_percentage_numbers_chars,
        TokenPreProcess.is_percentage_numbers,
        TokenPreProcess.is_multiple_smiley_in,
        TokenPreProcess.is_multiple_smiley,
        TokenPreProcess.is_one_char_fixable,

        # These need recursive handling
        TokenPreProcess.is_imp,
        TokenPreProcess.is_isp,
        TokenPreProcess.is_fmp,
        TokenPreProcess.is_fsp,
        TokenPreProcess.is_mssp,
        TokenPreProcess.is_msp,
        TokenPreProcess.is_midp,

        # Raw Punctuation
        TokenPreProcess.is_three_or_more,
        TokenPreProcess.is_roman_number,

        # Final Checks
        TokenPreProcess.is_multiple_emoticon,
        TokenPreProcess.is_non_latin,
        TokenPreProcess.is_num_char_sequence
    ]


class TokenProcessor:

    def __init__(self):
        pass

    @staticmethod
    def format_output(result, output_format):
        if output_format == 'tuple':
            return result
        elif output_format == 'list':
            return list(result)
        elif output_format == 'string':
            return f"{result[0]}\t{result[1]}"

    #@staticmethod
    #def process_token(token: str, output_format: str = 'tuple') -> tuple:
    #    for check in check_methods:
    #        result = check(token)
    #        if result:
    #            return TokenProcessor.format_output(result, output_format)

        # If no checks match, return "OOV" with the requested output format
    #    oov_result = (token, "OOV")
    #    return TokenProcessor.format_output(oov_result, output_format)
    @staticmethod
    def process_token(token: str, output_format: str = 'tuple') -> tuple:
        for check in check_methods:
            result = check(token)
            if result:
                return TokenProcessor.format_output(result, output_format)
        return TokenProcessor.format_output((token, "OOV"), output_format)




import re
import html
import unicodedata
# Tuples for character replacement
REPLACEMENTS_CHAR = [
    ("þ", "ş"), ("Þñ", "ı"), ("ð", "ğ"), ("ɪ", "ı"), ("ḡ", "ğ"), ("¤", "ğ"), ("а", "a"), ("ƒ", "a"),
    ("œ", "i"), ("ǧ", "ğ"), ("е", "e"), ("åÿ", "ş"), ("ș", "ş"), ("ɑ", "a"), ("о", "o"), ("í", "i"),
    ("Ã¶", "ö"), ("Ã¼", "ü"), ("Ã1/4", "ü"), ("Ã1/4", "ü"), ("Ã§", "ç"), ("Ä±", "ı"), ("›", "ı"), ("Ý", "İ"), ("Ã", "Ö"), ("Ð", "Ğ"), ("ġ", "Ş"),
    ("ÃŸ", "ş"), ("�", "ü"), ("…", "..."), ("»", ">"), ("«", "<"), ("s¸", "ş"), ("Ģ", "ş"), ("Þ", "Ş"), ("Ġ", "İ"),
    ("\\u011f", "ğ"), ("\\u00fc", "ü"), ("\\u0131", "ı"), ("\\u015f", "ş"), ("\\u00e7", "ç"), ("\\u00f6", "ö"),
    ("\\u0130", "İ"), ("\\u00dc", "Ü"), ("\\u015e", "Ş"), ("\\u00c7", "Ç"), ("&#252;", "ü"), ("ģ", "ş"), ("ä±", "ı"),
    ("õ", "ı"), ("ﬂ", "ş"), ("ä°", "i"), ("đ", "i"), ('\\u2022', ''),
    # Funny Char Problem
    ("ýº", "ış"), ("ºý", "şı"), ("üº", "üş"), ("ºü", "şü"), ("aº", "aş"), ("ºa", "şa"),
    ("uº", "uş"), ("ºu", "şu"), ("eº", "eş"), ("ºe", "şe"), ("ıº", "ış"), ("ºı", "şı"),
    ("iº", "iş"), ("ºi", "şi"), ("öº", "öş"), ("ºö", "şö"), ("oº", "oş"), ("ºo", "şo"),
    ("  ̊", "°"),
    ("Ä°", "İ"),
    ("ý", "ı"),
    # Fix Unicoed Chars
    ("і", "i"),
    # Fix Dashes - Need a better solution
    ("-", "-"), ("–", "-"), ("⁄", "/"), ("-", "-"), ("—-", "-"), ("﴾", "("), ("﴿", ")"),
    # punctuation_process.py line 178 - fails if not replaced
    # needs a better solution
    ("°°", "°")
]

# Tuples for html entities replacement
# Non-Breaking space &nbsp; is handled by removing.
REPLACEMENTS_HTML = [
    ("&amp;", "&"), ("&lt;", "<"), ("&gt;", ">"), ("&quot;", "\""), ("&nbsp;", ""), ("&Uuml;", "Ü"),
    ("&uuml;", "ü"), ("&Ouml;", "Ö"), ("&ouml;", "ö"), ("&Ccedil;", "Ç"), ("&ccedil;", "ç"),
    ("&Iuml;", "İ"), ("&iuml;", "i"), ("&ETH;", "Ğ"), ("&eth;", "ğ"), ("&THORN;", "Ş"), ("&thorn;", "ş"),
    ("&Auml;", "Ä"), ("&auml;", "ä"), ("&szlig;", "ß"), ("&rsquo;", "'"), ("&lsquo;", "‘"), ("&ndash;", "-"),
    ("&raquo;&raquo;", ">"), ("&lrm;", ""), ("&rlm;", ""), ("&raquo;", ">"), ("&laquo;", "<")
]

# Tuples for quotation mark replacement
REPLACEMENTS_QUOTE = [
    ("\"\"", "\""), (" ́", "'"), ("ʹ", "'"), ("ʺ", "\""),
    ("\"\"", "\""), ("“", "\""), ("”", "\""), ("’’", "\""), ("‘‘", "\""), ("‘’", "\""), ("‘’", "\""), ("", "\""), ("", "\""), ("““", "\""),
    ("''", "\""), ("", "\""), ("â", "'"), ("‘", "'"), ("’", "'"), ("", "'"), ("â€™", "'"), ("â€š", ","),
    ("’", "'"), ("’", "'"), ("’", "'"), ("\\u2018", "\""), ("\\u2019", "'"), ("˵", "'"), ("˶", "'"), ("'", "'"),
    ("\\u201c", "'"), ("\\u201d", "'"), ("ʼ", "'"), ("``", "'"), ("´", "'"), ("`", "'"), ("´", "'"), ("’", "'"), ("′′", "'"), ("′", "'"),
    ("‟", "'"), ("́", "'")
]

REPLACEMENTS_CONTROL_CHAR = [
    ("\u200a", ""), ("\u200b", ""), ("\u200c", ""),
    ("\u200d", ""), ("\u200e", ""), ("\u200f", "")
 ]


class CharFix:
    _compiled_pattern_char = re.compile("|".join(map(re.escape, dict(REPLACEMENTS_CHAR).keys())))
    _compiled_pattern_control = re.compile("|".join(map(re.escape, dict(REPLACEMENTS_CONTROL_CHAR).keys())))
    _compiled_pattern_html = re.compile("|".join(map(re.escape, dict(REPLACEMENTS_HTML).keys())))
    _compiled_pattern_quote = re.compile("|".join(map(re.escape, dict(REPLACEMENTS_QUOTE).keys())))

    @staticmethod
    def batch_replace(word: str, compiled_pattern, replacements: dict) -> str:
        return compiled_pattern.sub(lambda match: replacements[match.group(0)], word)

    @staticmethod
    def char_check(word: str) -> str:
        return CharFix.batch_replace(word, CharFix._compiled_pattern_char, dict(REPLACEMENTS_CHAR))

    @staticmethod
    def tr_lowercase(word: str) -> str:
        conversion = {'I': 'ı', 'İ': 'i'}
        for key, value in conversion.items():
            word = word.replace(key, value)
        return word.lower()

    @staticmethod
    def remove_unicode_controls(word: str) -> str:
        return CharFix.batch_replace(word, CharFix._compiled_pattern_control, dict(REPLACEMENTS_CONTROL_CHAR))

    @staticmethod
    def html_entity_replace(word: str) -> str:
        return CharFix.batch_replace(word, CharFix._compiled_pattern_html, dict(REPLACEMENTS_HTML))

    @staticmethod
    def fix_quote(word: str) -> str:
        return CharFix.batch_replace(word, CharFix._compiled_pattern_quote, dict(REPLACEMENTS_QUOTE))

    @staticmethod
    def remove_diacritics(word: str) -> str:
        normalized = unicodedata.normalize('NFD', word)
        return ''.join(c for c in normalized if not unicodedata.combining(c))

    @staticmethod
    def replace_diacritics(word: str) -> str:
        replacements = {"́": "'"}
        return ''.join(replacements.get(c, c) for c in word)

    @staticmethod
    def balance_quotes(word: str) -> str:
        # Handle empty word case
        if not word:
            return word

        # Count occurrences of quotes
        single_quote_count = word.count("'")
        double_quote_count = word.count("\"")

        # Ensure quotes are balanced
        if single_quote_count % 2 != 0:
            word = word.replace("'", "\"", 1)  # Replace the first unmatched single quote with double quote

        if double_quote_count % 2 != 0:
            word = word.replace("\"", "'", 1)  # Replace the first unmatched double quote with single quote

        return word

    @staticmethod
    def replace_all(word: str) -> str:
        word = CharFix.remove_unicode_controls(word)  # Remove control characters
        #word = CharFix.replace_diacritics(word)
        word = CharFix.fix_quote(word)
        word = CharFix.char_check(word)  # Apply character fixes
        word = CharFix.html_entity_replace(word)  # Replace HTML entities
        # word = CharFix.balance_quotes(word)
        return html.unescape(word)
        #return word

    @staticmethod
    def fix(word: str) -> str:
        return CharFix.replace_all(word)



import re
import string
from typing import Union, Optional

num_suffix = ["ıncı", "inci", "üncü", "uncu"]


class InnerPuncParser:
    def __init__(self):
        self.puncs = re.escape(string.punctuation)
        self.pattern = r'^(?P<initial>[%s]*)(?P<word>.*?)(?P<final>[%s]*)$' % (self.puncs, self.puncs)

    @classmethod
    def ord_number(cls, word: str) -> Optional[Union[bool, tuple]]:
        if "'" in word:
            word_parts = word.split("'")
            if word_parts[0].isdigit() and word_parts[1] in num_suffix:
                return word, "Ord_Number"



import re


class DateValidator:
    def __init__(self):
        pass

    @classmethod
    def is_valid_date_dd_mm_yyyy(cls, date_str: str) -> bool:
        match = re.match(r"^(0?[1-9]|[1-2][0-9]|3[0-1])([-/.:])(0?[1-9]|1[0-2])\2(\d{2,4})$", date_str)
        if not match:
            return False
        day, _, month, year = match.groups()
        day, month, year = map(int, [day, month, year])
        return cls._is_valid_day_month_year(day, month, year)

    @classmethod
    def is_valid_date_mm_dd_yyyy(cls, date_str: str) -> bool:
        match = re.match(r"^(0?[1-9]|1[0-2])([-/.:])(0?[1-9]|[1-2][0-9]|3[0-1])\2(\d{2,4})$", date_str)
        if not match:
            return False
        month, _, day, year = match.groups()
        month, day, year = map(int, [month, day, year])
        return cls._is_valid_day_month_year(day, month, year)

    @classmethod
    def _is_valid_day_month_year(cls, day: int, month: int, year: int) -> bool:
        if month in [1, 3, 5, 7, 8, 10, 12]:
            max_day = 31
        elif month in [4, 6, 9, 11]:
            max_day = 30
        elif month == 2:
            if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
                max_day = 29
            else:
                max_day = 28
        else:
            return False
        return 1 <= day <= max_day


class DateCheck:
    @staticmethod
    def is_date(date_str: str) -> bool:
        # Check if the entire string is a date
        return (
                DateValidator.is_valid_date_mm_dd_yyyy(date_str.strip()) or
                DateValidator.is_valid_date_dd_mm_yyyy(date_str.strip())
        )



"""
ts-tokenizer Command-Line Interface
Author: tanersezerr@gmail.com
"""

import sys
from ts_tokenizer.tokenizer import TSTokenizer

def main():
    # Check if input is piped via stdin
    if not sys.stdin.isatty():  # If input is coming from stdin (i.e., piped input)
        input_text = sys.stdin.read().strip()
        if not input_text:
            print("Error: No input received from stdin.")
            sys.exit(1)

        # Tokenize the input text from stdin
        try:
            TSTokenizer.ts_tokenize(input_text=input_text)
        except Exception as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)
    elif len(sys.argv) < 2:
        print("Usage: ts-tokenizer <filename> or pipe input via stdin (e.g., echo 'text' | ts-tokenizer)")
        sys.exit(1)
    else:
        # Tokenize from the provided file
        try:
            TSTokenizer.ts_tokenize()
        except Exception as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

if __name__ == '__main__':
    main()



from .data import LocalData


class SmileyParser:
    @classmethod
    def smiley_count(cls, text):
        smiley_count = 0
        for smiley in LocalData.smileys():
            smiley_count += text.count(smiley)
        return smiley_count

    @classmethod
    def smiley_tokenize(cls, text):
        tokenized = []
        i = 0
        while i < len(text):
            found_smiley = False
            for smiley in LocalData.smileys():
                if text[i:i+len(smiley)] == smiley:
                    tokenized.append(smiley)
                    i += len(smiley)
                    found_smiley = True
                    break
            if not found_smiley:
                i += 1
        return "\n".join(tokenized)

    @classmethod
    def consecutive_smiley(cls, text):
        if not text:
            return False

        found_smileys = []
        i = 0
        while i < len(text):
            for smiley in LocalData.smileys():
                if text[i:i + len(smiley)] == smiley:
                    found_smileys.append((i, smiley))
                    i += len(smiley) - 1  # Adjust index to continue after the current smiley
                    break
            i += 1

        if len(found_smileys) <= 1:
            return False

        for i in range(1, len(found_smileys)):
            # Check if the current smiley starts immediately after the previous one
            if found_smileys[i][0] != found_smileys[i - 1][0] + len(found_smileys[i - 1][1]):
                return False
        return True



import sys
import argparse
import multiprocessing
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor
from ts_tokenizer.token_handler import TokenProcessor, TokenPreProcess
from ts_tokenizer.char_fix import CharFix

# Tokenizer class definition
class TSTokenizer:

    @staticmethod
    def parse_arguments():
        """
        Parses command-line arguments.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-o", "--output", choices=["tokenized", "lines", "tagged", "tagged_lines"], default="tokenized",
            help="Specify the output format"
        )
        parser.add_argument("filename", help="Name of the file to process")
        parser.add_argument("-w", "--word", action="store_true", help="Enable CLI input mode")
        parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose mode")
        parser.add_argument("-j", "--jobs", type=int, help="Number of parallel workers", default=multiprocessing.cpu_count() - 1)
        return parser.parse_args()

    @staticmethod
    def tokenize_line(line, return_format):
        # First check if the line is an XML tag
        xml_tag = TokenPreProcess.is_xml(line)
        if xml_tag:
            # If the return format is tagged, output it as a tagged XML tag
            if return_format == 'tagged':
                return "\t".join(xml_tag)  # Return as XML_Tag with tag
            else:
                return line  # Just return the original line (XML tag) unchanged

        # Proceed with tokenization if the line is not XML
        processed_tokens = [TokenProcessor.process_token(token) for token in line.split() if token]
        flat_tokens = []
        for token_list in processed_tokens:
            if isinstance(token_list, list):
                flat_tokens.extend(token_list)
            else:
                flat_tokens.append(token_list)

        # Handle output formats

        # Handle output formats
        if return_format == 'tokenized':
            return '\n'.join([token[0] for token in flat_tokens])
        elif return_format == 'tagged':
            return '\n'.join([f"{token[0]}\t{token[1]}" for token in flat_tokens])
        elif return_format == 'lines':
            return [token[0] for token in flat_tokens]
        elif return_format == 'tagged_lines':
            return [(token[0], token[1]) for token in flat_tokens]  # Returns a list of tuples

    @staticmethod
    def ts_tokenize():
        args = TSTokenizer.parse_arguments()
        num_workers = args.jobs

        if args.word:
            word = sys.argv[-1]
            print(TSTokenizer.tokenize_line(word, args.output))
        else:
            # Process file input
            with open(args.filename, encoding='utf-8') as in_file:
                if args.verbose:
                    total_lines = sum(1 for _ in in_file)
                    in_file.seek(0)
                    pbar = tqdm(total=total_lines, desc="Processing File")
                else:
                    pbar = None

                with ProcessPoolExecutor(max_workers=num_workers) as executor:
                    batch_size = 100
                    batch = []

                    for line in in_file:
                        # Only strip spaces for non-XML tags
                        if not TokenPreProcess.is_xml(line):
                            line = CharFix.fix(line.strip())  # Strip whitespace and fix characters for non-XML lines
                        else:
                            line = CharFix.fix(line)  # Just apply CharFix without stripping spaces for XML tags

                        if line:  # Only process non-empty lines
                            batch.append(line)

                        if len(batch) >= batch_size:
                            futures = [executor.submit(TSTokenizer.tokenize_line, line, args.output) for line in batch]
                            for future in futures:
                                print(future.result())
                            batch = []  # Clear batch after processing

                        if pbar:
                            pbar.update(batch_size)

                    # Process remaining lines in the last batch
                    if batch:
                        futures = [executor.submit(TSTokenizer.tokenize_line, line, args.output) for line in batch]
                        for future in futures:
                            print(future.result())
                        if pbar:
                            pbar.update(len(batch))

                if pbar:
                    pbar.close()


if __name__ == "__main__":
    TSTokenizer.ts_tokenize()



import os

base_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(base_dir, 'data')

# Read lists for Emoticons, Smileys, Abbreviations and Valid Word List Extracted from TS Corpus

emoticons = set(line.strip() for line in open(os.path.join(data_dir, 'emoticons.txt')))
smileys = set(line.strip() for line in open(os.path.join(data_dir, 'smileys.txt')))
abbrs = set(line.strip() for line in open(os.path.join(data_dir, 'abbr_list.txt')))
word_list = set(line.strip() for line in open(os.path.join(data_dir, 'TS_Corpus_Turkish_Word_List.txt')))
exception_words = set(line.strip() for line in open(os.path.join(data_dir, 'exceptions.txt')))
eng_words = set(line.strip() for line in open(os.path.join(data_dir, 'eng_word_list.txt')))
domains = set(line.strip() for line in open(os.path.join(data_dir, 'domains.txt')))
currencies = set(line.strip() for line in open(os.path.join(data_dir, 'currency_symbols.txt')))
correction_mark = set(line.strip() for line in open(os.path.join(data_dir, 'corection_mark.txt')))


class LocalData:
    @staticmethod
    def emoticons():
        return emoticons

    @staticmethod
    def smileys():
        return smileys

    @staticmethod
    def abbrs():
        return abbrs

    @staticmethod
    def word_list():
        return word_list

    @staticmethod
    def correction_mark():
        return word_list

    @staticmethod
    def exception_words():
        return exception_words

    @staticmethod
    def eng_word_list():
        return eng_words

    @staticmethod
    def domains():
        return domains

    @staticmethod
    def currency_symbols():
        return currencies



from .data import LocalData

class EmoticonParser:

    @classmethod
    def emoticon_count(cls, word):
        emoticon_count = 0
        chars = list(word)
        chars = [char for char in chars if not char.isdigit() and not char.isalpha()]
        for char in chars:
            if char in LocalData.emoticons():
                emoticon_count += 1
        return emoticon_count


    @classmethod
    def emoticon_check(cls, word):
        emoticon_count = cls.emoticon_count(word)
        return emoticon_count

    @classmethod
    def emoticon_tokenize(cls, text):
        tokens = []
        word = ""
        for char in text:
            if char in LocalData.emoticons():
                if word:
                    tokens.append(word)
                    word = ""
                tokens.append(char)
            else:
                word += char
        if word:
            tokens.append(word)

        return "\n".join(tokens)



import re
import string
from .data import LocalData
from .punctuation_process import PuncMatcher
from .char_fix import CharFix
from .token_processor import TokenProcessor


FMP_exception_list = ["(!)", "...", "!!!"]


class ParseTokens:

    @classmethod
    def tokenize_xml(cls, word: str) -> str:
        return "".join(word)

    @classmethod
    def split_punctuation_and_chars(cls, word: str) -> list:
        patterns = ["...", "(!)", "!!!"] + list(LocalData.smileys())
        pattern_regex = '|'.join(map(re.escape, patterns))
        tokens = re.split('(' + pattern_regex + ')', word)
        return [token for token in tokens if token]

    @classmethod
    def tokenize_initial_quote(cls, word: str) -> str:
        word_list = list(word)
        index_list = [0, len(word)]
        for index in index_list:
            if index < len(word):
                word_list[index] = '\n' + word_list[index] + '\n'
        return ''.join(word_list).lstrip().rstrip()

    @classmethod
    def tokenize_in_quotes(cls, word: str) -> str:
        word_list = list(word)
        index_list = [0, len(word) - 1]
        for index in index_list:
            if index < len(word):
                word_list[index] = '\n' + word_list[index] + '\n'
        return ''.join(word_list).lstrip().rstrip()

    @classmethod
    def tokenize_isp(cls, word: str) -> str:
        if word and word[0] in string.punctuation:
            return word[0] + '\n' + word[1:]
        return word

    @classmethod
    def tokenize_imp(cls, word: str) -> str:
        word_list = list(word)
        index = 0
        if any(smiley in word for smiley in LocalData.smileys()):
            return '\n'.join(cls.split_punctuation_and_chars(word))
        if word.startswith("(!)"):
            return "\n".join(("(!)", word[3:]))

        else:
            while index < len(word_list):
                if word_list[index] in string.punctuation:
                    index += 1
                    word_list.insert(index, '\n')
                    index += 1
                else:
                    break
        return ''.join(word_list).strip()

    @classmethod
    def tokenize_fsp(cls, word: str) -> str:
        if word and word[-1] in string.punctuation:
            return word[:-1] + '\n' + word[-1]
        return word

    @classmethod
    def tokenize_msp(cls, word: str) -> str:
        word_list = list(word)
        index_list = [i for i, char in enumerate(word) if char in string.punctuation]
        for index in index_list:
            if 0 <= index < len(word):
                word_list[index] = f"\n{word_list[index]}\n"
        return ''.join(word_list).strip()

    @classmethod
    def tokenize_mssp(cls, word: str) -> list:
        # Strip first and last characters (they are punctuations) and recursively process the rest
        if len(word) > 2 and word[0] in string.punctuation and word[-1] in string.punctuation:
            inner_part = word[1:-1]  # The core token without the outer punctuations
            # Return the split punctuation with recursive processing for the inner part
            return [word[0]] + [TokenProcessor.process_token(inner_part, output='all')] + [word[-1]]
        return [word]  # Return the word as is if it doesn't match the MSSP pattern

    @classmethod
    def tokenize_fmp(cls, word: str) -> str:
        punc_count = PuncMatcher.punc_count(word)
        punc_positions = PuncMatcher.punc_pos(word)
        first_punc_pos = punc_positions[0]
        char_part = word[:first_punc_pos]
        punc_part = word[first_punc_pos:]

        if punc_part in LocalData.smileys():
            return f"{char_part}\n{punc_part}"
        elif punc_part in FMP_exception_list:
            return f"{char_part}\n{punc_part}"
        elif any(punc_part.startswith(exc) for exc in FMP_exception_list):
            return f"{char_part}\n" + "\n".join(cls.split_punctuation_and_chars(punc_part))
        elif len(punc_part) == punc_count:
            return f"{char_part}\n" + "\n".join(punc_part)

        # If the entire word is not punctuation
        if PuncMatcher.punc_count(word) != len(word):
            word_list = list(word)
            for index in punc_positions:
                if 0 < index < len(word):  # Avoid inserting newline at start
                    word_list[index] = '\n' + word_list[index]

            return ''.join(word_list).strip()

    @classmethod
    def tokenize_complex_punc(cls, word: str) -> str:
        # Exception case: one char + dot + one char + dot (e.g., "M.S.")
        if len(word) == 4 and all([word[0].isalpha(), word[1] == '.', word[2].isalpha(), word[3] == '.']):
            return word

        word_list = list(word)
        index_adjustment = 0
        for index in range(len(word)):
            if word[index] in string.punctuation and word[index] != "'":
                # Insert newline before and after the punctuation, adjusting index for the newly inserted characters
                word_list.insert(index + index_adjustment, '\n')
                index_adjustment += 1
                word_list.insert(index + index_adjustment + 1, '\n')
                index_adjustment += 1

        return ''.join(word_list).strip()

    @classmethod
    def tokenize_in_parenthesis(cls, word: str) -> str:
        if word in FMP_exception_list:
            return word
        elif word.startswith("(!)"):
            return ParseTokens.tokenize_fsp(word)
        else:
            word_list = list(word)
            index_list = [0, len(word) - 1]
            for index in index_list:
                if index < len(word):
                    word_list[index] = '\n' + word_list[index] + '\n'
            return ''.join(word_list).lstrip().rstrip()

    @classmethod
    def tokenize_mishyphenated(cls, word: str) -> str:
        word = word.replace("-", "")
        return word

    @classmethod
    def tokenize_inner_punc(cls, word: str) -> str:
        word = CharFix.fix(word)
        # Define exceptions
        exceptions = ["...", "(!)"]
        # If the word contains exactly one apostrophe not at the start or end
        if word.count("'") == 1 and not word.startswith("'") and not word.endswith("'"):
            return word

        # If the word contains exactly one punctuation character
        elif PuncMatcher.punc_count(word) == 1:
            # Get the position of the punctuation
            punc_pos = PuncMatcher.punc_pos(word)[0]
            # Split the word into three parts: before punctuation, punctuation, after punctuation
            first_part = word[:punc_pos]
            punc = word[punc_pos]
            second_part = word[punc_pos + 1:]
            # Return the parts separated by newlines
            return f"{first_part}\n{punc}\n{second_part}"

        elif PuncMatcher.punc_count(word) == 3:
            for exception in exceptions:
                if exception in word:
                    punc_pos = word.index(exception)
                    first_part = word[:punc_pos]
                    punc = exception
                    second_part = word[punc_pos + 3:]
                    return f"{first_part}\n{punc}\n{second_part}"

        # Default case: tokenize the word normally
        result = ""
        start = 0
        index_list = [i for i, char in enumerate(word) if char in string.punctuation]
        for index in index_list:
            result += word[start:index] + '\n'
            result += word[index] + '\n'
            start = index + 1
        result += word[start:]
        return result.strip().replace("\n\n", "\n")

    @classmethod
    def tokenize_date(cls, word: str) -> str:
        if word[0] in string.punctuation:
            return "\n".join([word[0], word[1:]])  # Split first punctuation and rest of the word
        elif word[-1] in string.punctuation:
            return "\n".join([word[:-1], word[-1]])  # Split rest of the word and last punctuation
        return word  # Return word as is if no surrounding punctuation

    @classmethod
    def tokenize_one_char_fixed(cls, word: str) -> str:
        extra_chars = ["¬", "-"]
        for extra in extra_chars:
            fixed_word = word.replace(extra, "")
            return fixed_word



from setuptools import setup, find_packages

setup(
    name='ts_tokenizer',
    version='0.1.5',
    packages=find_packages(),
    package_data={
    'ts_tokenizer': ['data/*.txt'],
    },
    install_requires=[
        'tqdm~=4.66.4'
    ],
    entry_points={
        'console_scripts': [
            'ts-tokenizer=ts_tokenizer.cli:main',
        ],
    },
    author='Taner Sezer',
    author_email='tanersezerr@gmail.com',
    description='TS Tokenizer is a rule-based tokenizer'
                'specifically designed for processing Turkish text.'
                'It provides functionalities to split text into tokens'
                'following the grammatical and linguistic rules of the Turkish language.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/tanerim/ts_tokenizer',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    repository='https://github.com/tanerim/ts_tokenizer'
)



import re
import string
from typing import Union, Optional
from .char_fix import CharFix
from .data import LocalData
from .date_check import DateCheck


puncs = re.escape(string.punctuation)

extra_puncs = ["–", "'", "°", "—"]
for p in extra_puncs:
    puncs = puncs + p

PuncPattern = r'^(?P<initial>[%s]*)(?P<word>.*?)(?P<final>[%s]*)$' % (puncs, puncs)


def fix_tr_lowercase(word):
    conversion = {'I': 'ı', 'İ': 'i'}
    for key, value in conversion.items():
        word = word.replace(key, value)
    return word.lower()


class PuncMatcher:

    @classmethod
    def punc_count(cls, word: str) -> int:
        return sum(1 for char in word if char in string.punctuation)

    @classmethod
    def punc_pos(cls, word: str) -> list:
        return [i for i, char in enumerate(word) if char in string.punctuation]

    @classmethod
    def find_punctuation(cls, word: str) -> Optional[Union[bool, str]]:
        match = re.match(PuncPattern, word)
        if not match:
            return None

        initial_punc = match.group('initial')
        final_punc = match.group('final')

        if initial_punc and final_punc:
            if len(initial_punc) == 1 and len(final_punc) == 1:
                return "MSSP"  # Multi-Side Single Punctuation
            else:
                return "MSP"  # Multi-Side Punctuation
        elif initial_punc:
            if len(initial_punc) == 1:
                return "ISP"  # Initial_Single_Punc
            else:
                return "IMP"  # Initial_Multi_Punc
        elif final_punc:
            if len(final_punc) == 1 and PuncMatcher.punc_count(word) == 1:
                return "FSP"  # Final_Single_Punc
            else:
                return "FMP"  # Final_Multi_Punc
        return None

    @classmethod
    def inner_punctuation(cls, word: str) -> Optional[Union[bool, str]]:
        # Count the punctuations in the word
        punc_count = PuncMatcher.punc_count(word)

        if PuncMatcher.hyphen_in(word):
            return PuncMatcher.hyphenated(word)

        # Check if the word has no punctuation at the start and end, but punctuation inside
        if punc_count == 1 and word[0] not in string.punctuation and word[-1] not in string.punctuation and "-" not in word:
            return "Inner_Single_Punc"

        # Check if there are multiple punctuations inside the word and no punctuation at the start or end
        elif punc_count >= 2:
            # Check if there are punctuations within the middle of the word
            inner_punctuation = any(char in string.punctuation for char in word[1:-1])
            # Date range (assumed implementation of is_date to check valid date)
            parts = word.split("-")
            if all(DateCheck.is_date(part) for part in parts):
                return "Date_Range"
            #else:
            if inner_punctuation:
                return "Inner_Multiple_Punc"
        elif "°" in word:
            if word.count("°") == 1 and word[-1] == "°":
                return "Celcius"
            elif word.count("°") == 2 and word[-1] == "°":
                return "Celcius_Range"
            else:
                return "Coordinate"

        return None

    @classmethod
    def apostrophed(cls, word: str) -> Optional[Union[bool, str]]:
        match = re.match(r"\b\w+'[a-zA-Z]+\b", word)
        if match:
            punc_positions = cls.punc_pos(word)
            if cls.punc_count(word) == 1 and punc_positions and "'" in word:
                return "apostrophes_in"
        return None

    @staticmethod
    def hyphenated(word):
        # Check for hyphenated words where both parts exist in the word list
        if ("-" in word and word.count("-") == 1 and word[0] != "-" and word[-1] != "-"
                and not any(c in string.punctuation for c in word if c != "-")):
            parts = word.split("-")
            if all(fix_tr_lowercase(part) in LocalData.word_list() for part in parts):
                return "Hyphenated"  # Valid hyphenated word

    @staticmethod
    def hyphen_in(word):
        parts = word.split("-")
        cleaned_word = CharFix.fix(word)
        hyphen_count = word.count("-")

        if "-" not in word:
            return None

        if hyphen_count >= 2:

            if word[0] != "-" and word[-1] != "-":
                return "Multi_Hyphenated"

        elif hyphen_count == 1:
            if all(part.isdigit() for part in parts):
                tag = "Year_Range" if all(len(part) == 4 for part in parts) else "Numeric_Hyphenated"
                return tag

            elif PuncMatcher.hyphenated(word):
                return "Hyphenated"

            elif word[0] == "-" and word[-1] != "-":
                # You could handle specific cases, like if you want to treat these words differently
                return "Hyphen_Initial"

            elif word[0] != "-" and word[-1] == "-":
                # You could handle specific cases, like if you want to treat these words differently
                return "Hyphen_Final"

            elif any(part.isdigit() or any(c.isalpha() for c in part) for part in parts):
                return "Single_Hyphenated"

        elif word.count("-") >= 2:
            return word, cleaned_word, "Multi_Hyphens"

        return None


class PuncTagCheck:

    @classmethod
    def punc_tag_check(cls, word: str) -> tuple:
        word = CharFix.fix(word)
        punc_count = PuncMatcher.punc_count(word)
        punc_loc = PuncMatcher.punc_pos(word)
        hyphen_check = PuncMatcher.hyphen_in(word)
        inner_punc = PuncMatcher.inner_punctuation(word)
        apostrophe = PuncMatcher.apostrophed(word)
        complex_punc = PuncMatcher.find_punctuation(word)

        if hyphen_check:
            return PuncMatcher.hyphen_in(word), word, punc_count, punc_loc

        if apostrophe:
            return "Apostrophe", word, punc_count, punc_loc

        if complex_punc:
            s_word = ""
            if complex_punc == "MSSP":
                s_word = (word[0], word[1:-1], word[-1])
            elif complex_punc == "MSP":
                s_word = (word[0], word[1:-1], word[-1])
            elif complex_punc == "ISP":
                s_word = (word[0], word[1:])
            elif complex_punc == "FSP":
                s_word = ("FSP", word[0:-1], word[-1])
            elif complex_punc == "FMP":
                p_start = punc_loc[0]
                s_word = (word[0:p_start], word[p_start:])
            elif complex_punc == "IMP":
                p_end = punc_loc[-1] + 1
                s_word = (word[0:p_end], word[p_end:])
            return complex_punc, word, s_word, punc_count, punc_loc

        if inner_punc:
            return inner_punc, word, punc_count, punc_loc

        return word, punc_count, punc_loc



from ts_tokenizer.token_handler import TokenProcessor

class oov_parser:
    def __init__(self):
        pass

    @staticmethod
    def isp_parser(word: str) -> tuple:
        if word[0] == ",":
            return TokenProcessor.process_token(word[1:])


word = ",yeniden,eski"

print(oov_parser.isp_parser(word))